package edu.wpi.appjudicator.ssl.sslSocketChannel;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.security.SecureRandom;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLEngine;
import javax.net.ssl.SSLEngineResult;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLSession;

import android.os.SystemClock;
import android.util.Log;

import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Packet;

/**
 * An SSL/TLS client that connects to a server using its IP address and port.
 * <p/>
 * After initialization of a {@link NioSslClient} object, {@link NioSslClient#connect()} should be called,
 * in order to establish connection with the server.
 * <p/>
 * When the connection between the client and the object is established, {@link NioSslClient} provides
 * a public write and read method, in order to communicate with its peer. 
 *
 * @author <a href="mailto:alex.a.karnezis@gmail.com">Alex Karnezis</a>
 */


public class NioSslClient extends NioSslPeer {
    //NOTE: This is a workaround
    public enum TLS_CLient_State {
        STATE_CREATED,
        STATE_SENT_CLIENT_HELLO,
        STATE_RECEIVED_SERVER_HELLO,
        STATE_SENT_CLIENT_KEY_EXCHANGE,
        STATE_ENCRYPTED_COMMUNICATION,
        STATE_CLOSED_COMMUNICATION
    }
    public TLS_CLient_State cLient_state;

    private static final String TAG = "NioSslClient";

    /**
     * The remote address of the server this client is configured to connect to.
     */
	private String remoteAddress;

	/**
	 * The port of the server this client is configured to connect to.
	 */
	private int port;

	/**
	 * The engine that will be used to encrypt/decrypt data between this client and the server.
	 */
    private SSLEngine engine;

    /**
     * The socket channel that will be used as the transport link between this client and the server.
     */
    public SocketChannel socketChannel;


    /**
     * Initiates the engine to run as a client using peer information, and allocates space for the
     * buffers that will be used by the engine.
     *
     * @param protocol The SSL/TLS protocol to be used. Java 1.6 will only run with up to TLSv1 protocol. Java 1.7 or higher also supports TLSv1.1 and TLSv1.2 protocols.
     * @param remoteAddress The IP address of the peer.
     * @param port The peer's port that will be used.
     * @throws Exception
     */
    public NioSslClient(String protocol, String remoteAddress, int port) throws Exception  {
        //Log.d(TAG, "Created NioSslClient");
        if(remoteAddress.charAt(0) == '/') {
            this.remoteAddress = remoteAddress.substring(1);
        } else {
            this.remoteAddress = remoteAddress;
        }

        this.cLient_state = TLS_CLient_State.STATE_CREATED;
    	this.port = port;

        SSLContext context = SSLContext.getInstance(protocol);
        context.init(null, null, null);
        engine = context.createSSLEngine(remoteAddress, port);
        engine.setUseClientMode(true);

        SSLSession session = engine.getSession();
        myAppData = ByteBuffer.allocate(1024);
        myNetData = ByteBuffer.allocate(session.getPacketBufferSize());

        peerAppData = ByteBuffer.allocate(1024);
        peerNetData = ByteBuffer.allocate(session.getPacketBufferSize());

        myAppData.clear();
        myNetData.clear();
        
        peerAppData.clear();
        peerNetData.clear();

        socketChannel = SocketChannel.open();
        socketChannel.configureBlocking(false);
    }

    /**
     * Opens a socket channel to communicate with the configured server and tries to complete the handshake protocol.
     *
     * @return True if client established a connection with the server, false otherwise.
     * @throws Exception
     */

    /**
     * Implements the handshake protocol between two peers, required for the establishment of the SSL/TLS connection.
     * During the handshake, encryption configuration information - such as the list of available cipher suites - will be exchanged
     * and if the handshake is successful will lead to an established SSL/TLS session.
     * <p>
     * <p/>
     * A typical handshake will usually contain the following steps:
     *
     * <ul>
     *   <li>1. wrap:     ClientHello</li>
     *   <li>2. unwrap:   ServerHello/Cert/ServerHelloDone</li>
     *   <li>3. wrap:     ClientKeyExchange</li>
     *   <li>4. wrap:     ChangeCipherSpec</li>
     *   <li>5. wrap:     Finished</li>
     *   <li>6. unwrap:   ChangeCipherSpec</li>
     *   <li>7. unwrap:   Finished</li>
     * </ul>
     * <p/>
     * Handshake is also used during the end of the session, in order to properly close the connection between the two peers.
     * A proper connection close will typically include the one peer sending a CLOSE message to another, and then wait for
     * the other's CLOSE message to close the transport link. The other peer from his perspective would read a CLOSE message
     * from his peer and then enter the handshake procedure to send his own CLOSE message as well.
     *
     * @param socketChannel - the socket channel that connects the two peers.
     * @param engine        - the engine that will be used for encryption/decryption of the data exchanged with the other peer.
     * @return True if the connection handshake was successful or false if an error occurred.
     * @throws IOException - if an error occurs during read/write to the socket channel.
     */
    protected boolean doHandshake(SocketChannel socketChannel, SSLEngine engine) throws IOException {

        //Log.d(TAG, "About to do handshake...");

        SSLEngineResult result;
        SSLEngineResult.HandshakeStatus handshakeStatus;

        // NioSslPeer's fields myAppData and peerAppData are supposed to be large enough to hold all message data the peer
        // will send and expects to receive from the other peer respectively. Since the messages to be exchanged will usually be less
        // than 16KB long the capacity of these fields should also be smaller. Here we initialize these two local buffers
        // to be used for the handshake, while keeping client's buffers at the same size.
        int appBufferSize = engine.getSession().getApplicationBufferSize();
        ByteBuffer myAppData = ByteBuffer.allocate(appBufferSize);
        ByteBuffer peerAppData = ByteBuffer.allocate(appBufferSize);
        myNetData.clear();
        peerNetData.clear();

        handshakeStatus = engine.getHandshakeStatus();
        while (handshakeStatus != SSLEngineResult.HandshakeStatus.FINISHED && handshakeStatus != SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING && cLient_state != TLS_CLient_State.STATE_ENCRYPTED_COMMUNICATION) {
            switch (handshakeStatus) {
                case NEED_UNWRAP:
                    int bytes_read = socketChannel.read(peerNetData);
                    if (bytes_read < 0) {
                        if (engine.isInboundDone() && engine.isOutboundDone()) {
                            return false;
                        }
                        try {
                            engine.closeInbound();
                        } catch (SSLException e) {
                            Log.e(TAG, "This engine was forced to close inbound, without having received the proper SSL/TLS close notification message from the peer, due to end of stream.");
                        }
                        engine.closeOutbound();
                        // After closeOutbound the engine will be set to WRAP state, in order to try to send a close message to the client.
                        handshakeStatus = engine.getHandshakeStatus();
                        break;
                    } else if (bytes_read > 0) {
                        while (handshakeStatus == SSLEngineResult.HandshakeStatus.NEED_UNWRAP) {
                            peerNetData.flip();
                            try {
                                result = engine.unwrap(peerNetData, peerAppData);
                                peerNetData.compact();
                                handshakeStatus = result.getHandshakeStatus();
                            } catch (SSLException sslException) {
                                Log.e(TAG, "A problem was encountered while processing the data that caused the SSLEngine to abort. Will try to properly close connection...");
                                engine.closeOutbound();
                                handshakeStatus = engine.getHandshakeStatus();
                                break;
                            }
                            switch (result.getStatus()) {
                                case OK:
                                    break;
                                case BUFFER_OVERFLOW:
                                    // Will occur when peerAppData's capacity is smaller than the data derived from peerNetData's unwrap.
                                    peerAppData = enlargeApplicationBuffer(engine, peerAppData);
                                    break;
                                case BUFFER_UNDERFLOW:
                                    // Will occur either when no data was read from the peer or when the peerNetData buffer was too small to hold all peer's data.
                                    peerNetData = handleBufferUnderflow(engine, peerNetData);
                                    break;
                                case CLOSED:
                                    if (engine.isOutboundDone()) {
                                        return false;
                                    } else {
                                        engine.closeOutbound();
                                        handshakeStatus = engine.getHandshakeStatus();
                                        break;
                                    }
                                default:
                                    throw new IllegalStateException("Invalid SSL status: " + result.getStatus());
                            }
                            handshakeStatus = engine.getHandshakeStatus();
                        }
                        //NOTE: This is a workaround
                        if(this.cLient_state == TLS_CLient_State.STATE_SENT_CLIENT_HELLO) {
                            this.cLient_state = TLS_CLient_State.STATE_RECEIVED_SERVER_HELLO;
                        } else if (this.cLient_state == TLS_CLient_State.STATE_SENT_CLIENT_KEY_EXCHANGE) {
                            this.cLient_state = TLS_CLient_State.STATE_ENCRYPTED_COMMUNICATION;
                        }
                    }
                    break;
                case NEED_WRAP:
                    myNetData.clear();
                    myAppData.clear();
                    try {
                        result = engine.wrap(myAppData, myNetData);
                        handshakeStatus = engine.getHandshakeStatus();
                    } catch (SSLException sslException) {
                        Log.e(TAG, "A problem was encountered while processing the data that caused the SSLEngine to abort. Will try to properly close connection...");
                        engine.closeOutbound();
                        handshakeStatus = engine.getHandshakeStatus();
                        break;
                    }
                    switch (result.getStatus()) {
                        case OK:
                            myNetData.flip();
                            while (myNetData.hasRemaining() ){//&& handshakeStatus != HandshakeStatus.FINISHED) {
                                //TLS_Packet p = new TLS_Packet(myNetData.array());
                                socketChannel.write(myNetData);
                                //NOTE: This is a workaround
                                if(this.cLient_state == TLS_CLient_State.STATE_CREATED) {
                                    this.cLient_state = TLS_CLient_State.STATE_SENT_CLIENT_HELLO;
                                } else if (this.cLient_state == TLS_CLient_State.STATE_RECEIVED_SERVER_HELLO) {
                                    this.cLient_state = TLS_CLient_State.STATE_SENT_CLIENT_KEY_EXCHANGE;
                                }
                                handshakeStatus = engine.getHandshakeStatus();
                            }
                            break;
                        case BUFFER_OVERFLOW:
                            // Will occur if there is not enough space in myNetData buffer to write all the data that would be generated by the method wrap.
                            // Since myNetData is set to session's packet size we should not get to this point because SSLEngine is supposed
                            // to produce messages smaller or equal to that, but a general handling would be the following:
                            myNetData = enlargePacketBuffer(engine, myNetData);
                            break;
                        case BUFFER_UNDERFLOW:
                            throw new SSLException("Buffer underflow occured after a wrap. I don't think we should ever get here.");
                        case CLOSED:
                            try {
                                myNetData.flip();
                                while (myNetData.hasRemaining()) {
                                    socketChannel.write(myNetData);
                                }
                                // At this point the handshake status will probably be NEED_UNWRAP so we make sure that peerNetData is clear to read.
                                peerNetData.clear();
                            } catch (Exception e) {
                                Log.e(TAG, "Failed to send server's CLOSE message due to socket channel's failure.");
                                handshakeStatus = engine.getHandshakeStatus();
                            }
                            break;
                        default:
                            throw new IllegalStateException("Invalid SSL status: " + result.getStatus());
                    }
                    break;
                case NEED_TASK:
                    Runnable task;
                    while ((task = engine.getDelegatedTask()) != null) {
                        executor.execute(task);
                    }
                    handshakeStatus = engine.getHandshakeStatus();
                    break;
                case FINISHED:
                    break;
                case NOT_HANDSHAKING:
                    break;
                default:
                    throw new IllegalStateException("Invalid SSL status: " + handshakeStatus);
            }
        }


        isOpen = cLient_state == TLS_CLient_State.STATE_ENCRYPTED_COMMUNICATION;
        //Log.d(TAG, "Handshake Complete");
        return true;

    }

    public boolean connect() throws Exception {
       // Log.d(TAG, "Connecting to server...");
        if (!socketChannel.isConnected()) {
            socketChannel.connect(new InetSocketAddress(remoteAddress, port));
        }

        int attempts = 0;
    	while (!socketChannel.finishConnect() && attempts < 10){
            SystemClock.sleep(500);
            attempts += 1;
    	} if (attempts >= 10) {
    	    throw new RuntimeException("NioSSLCLient Timed Out");
        }

    	SSLEngineResult.HandshakeStatus handshakeStatus = engine.getHandshakeStatus();
        if (handshakeStatus == SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING && this.cLient_state != TLS_CLient_State.STATE_ENCRYPTED_COMMUNICATION)
        {
            engine.beginHandshake();
        }
    	return  doHandshake(socketChannel, engine);
    }

    /**
     * Public method to send a message to the server.
     *
     * @param message - message to be sent to the server.
     * @throws IOException if an I/O error occurs to the socket channel.
     */
    public void write(byte[] message) throws IOException {
        write(socketChannel, engine, message);
    }

    /**
     * Public method to send a message to the server.
     *
     * @param message - message to be sent to the server.
     * @throws IOException if an I/O error occurs to the socket channel.
     */
    public void write(String message) throws IOException {
        write(socketChannel, engine, message);
    }

    /**
     * Implements the write method that sends a message to the server the client is connected to,
     * but should not be called by the user, since socket channel and engine are inner class' variables.
     * {@link NioSslClient#write(String)} should be called instead.
     *
     * @param message - message to be sent to the server.
     * @param engine - the engine used for encryption/decryption of the data exchanged between the two peers.
     * @throws IOException if an I/O error occurs to the socket channel.
     */
    protected void write(SocketChannel socketChannel, SSLEngine engine, byte[] message) throws IOException {

        //Log.d(TAG,"About to write to the server...");

        myAppData.clear();
        myAppData.put(message);
        myAppData.flip();
        while (myAppData.hasRemaining()) {
            // The loop has a meaning for (outgoing) messages larger than 16KB.
            // Every wrap call will remove 16KB from the original message and send it to the remote peer.
            myNetData.clear();
            SSLEngineResult result = engine.wrap(myAppData, myNetData);
            switch (result.getStatus()) {
                case OK:
                    myNetData.flip();
                    while (myNetData.hasRemaining()) {
                        socketChannel.write(myNetData);
                    }
                    //Log.d(TAG,"Message sent to the server: " + message);
                    break;
                case BUFFER_OVERFLOW:
                    myNetData = enlargePacketBuffer(engine, myNetData);
                    break;
                case BUFFER_UNDERFLOW:
                    throw new SSLException("Buffer underflow occured after a wrap. I don't think we should ever get here.");
                case CLOSED:
                    closeConnection(socketChannel, engine);
                    return;
                default:
                    throw new IllegalStateException("Invalid SSL status: " + result.getStatus());
            }
        }

    }

    /**
     * Implements the write method that sends a message to the server the client is connected to,
     * but should not be called by the user, since socket channel and engine are inner class' variables.
     * {@link NioSslClient#write(String)} should be called instead.
     *
     * @param message - message to be sent to the server.
     * @param engine - the engine used for encryption/decryption of the data exchanged between the two peers.
     * @throws IOException if an I/O error occurs to the socket channel.
     */
    @Override
    protected void write(SocketChannel socketChannel, SSLEngine engine, String message) throws IOException {

       //Log.d(TAG,"About to write to the server...");

        myAppData.clear();
        myAppData.put(message.getBytes());
        myAppData.flip();
        while (myAppData.hasRemaining()) {
            // The loop has a meaning for (outgoing) messages larger than 16KB.
            // Every wrap call will remove 16KB from the original message and send it to the remote peer.
            myNetData.clear();
            SSLEngineResult result = engine.wrap(myAppData, myNetData);
            switch (result.getStatus()) {
            case OK:
                myNetData.flip();
                while (myNetData.hasRemaining()) {
                    socketChannel.write(myNetData);
                }
               Log.d(TAG,"Message sent to the server: " + message);
                break;
            case BUFFER_OVERFLOW:
                myNetData = enlargePacketBuffer(engine, myNetData);
                break;
            case BUFFER_UNDERFLOW:
                throw new SSLException("Buffer underflow occured after a wrap. I don't think we should ever get here.");
            case CLOSED:
                closeConnection(socketChannel, engine);
                return;
            default:
                throw new IllegalStateException("Invalid SSL status: " + result.getStatus());
            }
        }

    }

    /**
     * Public method to try to read from the server.
     *
     * @throws Exception
     */
    public int read(byte[] decr_data) throws Exception {
        return read(socketChannel, engine, decr_data);
    }

    /**
     * Will wait for response from the remote peer, until it actually gets something.
     * Uses {@link SocketChannel#read(ByteBuffer)}, which is non-blocking, and if
     * it gets nothing from the peer, waits for {@code waitToReadMillis} and tries again.
     * <p/>

     * @param engine - the engine used for encryption/decryption of the data exchanged between the two peers.
     * @throws Exception
     */
    protected int read(SocketChannel socketChannel, SSLEngine engine, byte[] decr_data) throws Exception  {

       //Log.d(TAG,"About to read from the server...");

        peerNetData.clear();

        int bytesRead = socketChannel.read(peerNetData);
        if (bytesRead > 0) {
            peerAppData.clear();
            peerNetData.flip();
            while (peerNetData.hasRemaining()) {
                SSLEngineResult result = engine.unwrap(peerNetData, peerAppData);
                switch (result.getStatus()) {
                case OK:
                    peerAppData.flip();
                    System.arraycopy(peerAppData.array(), peerAppData.arrayOffset(), decr_data, 0, result.bytesProduced());
                   //Log.d(TAG,"Server response: " + new String(peerAppData.array()));
                    break;
                case BUFFER_OVERFLOW:
                    peerAppData = enlargeApplicationBuffer(engine, peerAppData);
                    break;
                case BUFFER_UNDERFLOW:
                    peerNetData = handleBufferUnderflow(engine, peerNetData);
                    break;
                case CLOSED:
                    closeConnection(socketChannel, engine);
                    break;
                default:
                    throw new IllegalStateException("Invalid SSL status: " + result.getStatus());
                }
            }

        } else if (bytesRead < 0) {
            handleEndOfStream(socketChannel, engine);
        }
        return bytesRead;
    }

    /**
     * Should be called when the client wants to explicitly close the connection to the server.
     *
     * @throws IOException if an I/O error occurs to the socket channel.
     */
    public void shutdown() throws IOException {
       //Log.d(TAG,"About to close connection with the server...");
        engine.closeOutbound();
        closeConnection(socketChannel, engine);
        executor.shutdown();
    }

}
