package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import org.pcap4j.util.LazyValue;

import java.io.Serializable;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Record_Layer implements Serializable {
    private final LazyValue<byte[]> rawDataCache;

    public byte content_type;
    byte version_major;
    byte version_minor;
    byte[] length;
    public TLS_Sublayer payload;

    public TLS_Record_Layer(){

        rawDataCache = new LazyValue<byte[]>(
            new LazyValue.BuildValueCommand<byte[]>() {
                @Override
                public byte[] buildValue() {
                    return buildRawData();
                }
            }
    );;}

    public TLS_Record_Layer(TLS_Record_Layer_Builder b)  {
        this.content_type = b.content_type;
        this.version_major = b.version_major;
        this.version_minor = b.version_minor;
        this.length = new byte[b.length.length];
        System.arraycopy(b.length, 0, this.length, 0, this.length.length);


        try {
            this.payload = b.payload_builder.build();
        }
        catch (Exception e) {

        }

        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                }
        );;
    }

    public TLS_Record_Layer(byte[] raw_bytes, int start, int end) {
        int record_length = end - start + 1;
        if (record_length < 5) {
            throw new IllegalArgumentException("TLS Record Layer is too short");
        }
        content_type = raw_bytes[start];
        version_major = raw_bytes[start + 1];
        version_minor = raw_bytes[start + 2];

        length = new byte[2];
        System.arraycopy (raw_bytes, start + 3, length, 0, 2);

        try {
            TLS_Sublayer_Factory sublayer_factory = get_sublayer_factory();
            payload = sublayer_factory.build(raw_bytes, start + 5, SSL_Helper.convert_to_int(length, 0, 2));
        } catch (IllegalStateException e) {
            throw e;
        }
        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                }
        );;
    }

    public TLS_Sublayer_Factory get_sublayer_factory() {
        TLS_Enum_Content_Type c_type = TLS_Enum_Content_Type.getInstance(content_type);

        if (c_type.equals(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_HANDSHAKE)) {
            return new TLS_Handshake_Sublayer_Factory();
        } else if (c_type.equals(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_CHANGE_CYPHER_SPEC)) {
            return new TLS_Change_Cipher_Spec_Factory();
        } else if (c_type.equals(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_APPLICATION_DATA)){
            return new TLS_Application_Data_Factory();
        } else if (c_type.equals(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_ALERT)) {
            return new TLS_Alert_Factory();
        }
            else {
            throw new IllegalStateException("Not a valid sublayer type" + new StringBuilder().append(c_type).toString());
        }

    }

    public int getLength() {
        return SSL_Helper.convert_to_int(length, 0, 2);
    }

    public int getTotalSize() {
        return getLength() + 5;
    }


    public byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];

        rd[0] = content_type;
        rd[1] = version_major;
        rd[2] = version_minor;
        System.arraycopy(length, 0, rd, 3, 2);

        System.arraycopy(payload.getRawData(), 0, rd, 5, payload.getTotalSize());

        return rd;
    }

    public byte[] getRawData() {
        byte[] rawData = rawDataCache.getValue();

        byte[] copy = new byte[rawData.length];
        System.arraycopy(rawData, 0, copy, 0, copy.length);
        return copy;
    }

    public static class TLS_Record_Layer_Builder implements TLS_Builder<TLS_Record_Layer>{
        byte content_type;
        byte version_major;
        byte version_minor;
        byte[] length;
        public TLS_Builder<TLS_Sublayer> payload_builder;

        public TLS_Record_Layer_Builder() {
            this.content_type = 0;
            this.version_major = 0;
            this.version_minor = 0;
            this.length = null;
            this.payload_builder = null;
        }

        public TLS_Record_Layer_Builder(byte content_type) {
            this.content_type = content_type;
            this.version_major = 0x03;
            this.version_minor = 0x03;
            this.length = null;
            this.payload_builder = null;
        }

        public TLS_Record_Layer_Builder(TLS_Record_Layer r)  {
            this.content_type = r.content_type;
            this.version_major = r.version_major;
            this.version_minor = r.version_minor;
            this.length = new byte[r.length.length];
            System.arraycopy(r.length, 0, this.length, 0, this.length.length);
            TLS_Sublayer_Builder_Factory sublayer_builder_factory = get_sublayer_builder_factory(r.payload);
            this.payload_builder = sublayer_builder_factory.build(r.payload);

        }

        public TLS_Sublayer_Builder_Factory get_sublayer_builder_factory(TLS_Sublayer a_payload) {
            Class sublayer_type = a_payload.getClass();

            if (TLS_Handshake_Sublayer.class.isAssignableFrom(sublayer_type)) {
                return new TLS_Handshake_Sublayer.TLS_Handshake_Sublayer_Builder_Factory();
            } else {
                throw new IllegalStateException("Not a valid sublayer type");
            }
        }
        public TLS_Record_Layer build() {
            return new TLS_Record_Layer(this);
        }

        @Override
        public void fixSizeFields() {
            if(this.length == null) {
                this.length = new byte[2];
            }
            int sub_size = this.payload_builder.fixAndGetTotalSize();
            SSL_Helper.revert_to_bytes(sub_size, this.length, 0, this.length.length);
        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return SSL_Helper.convert_to_int(this.length, 0, this.length.length) + this.length.length + 3;
        }
    }
}
